from .TokenSpan import TokenSpan
from .remove_list_duplicates import remove_list_duplicates
from .Entity import Entity

class NounChunk(TokenSpan):

	def __init__(self, obj, document):
		super().__init__(obj=obj, document=document)
		for token in self.tokens:
			token._noun_chunk = self

	@property
	def child_entities(self):
		"""
		:rtype: list[Entity]
		"""
		return [e for e in self.document.entities if e.start >= self.start and e.end <= self.end]

	@property
	def parent_entities(self):
		"""
		:rtype: list[Entity]
		"""
		return [e for e in self.document.entities if e.start <= self.start and e.end >= self.end]

	@property
	def entities(self):
		"""
		:rtype: list[Entity]
		"""
		return remove_list_duplicates(self.child_entities + self.parent_entities)

	@property
	def id(self):
		return (self.document.id, 'noun_chunk', self.start, self.end)

	def graph_str(self):
		if self.entities:
			return f"{self}\n({str(self.main_entity_type).replace('_', ' ')})"
		else:
			return f'{self}'

	@property
	def entity_types(self):
		return [entity.entity_type for entity in self.entities]

	@property
	def main_entity_type(self):
		return sorted(self.entity_types)[0]


from .DocumentObject import DocumentObject

from spacy import tokens

class TokenSpan(DocumentObject):

	@property
	def obj(self):
		"""
		:rtype: tokens.Span
		"""
		return self._obj

	@property
	def start(self):
		"""
		:rtype: int
		"""
		return self.obj.start

	@property
	def end(self):
		"""
		:rtype: int
		"""
		return self.obj.end

	@property
	def id(self):
		return (self.document.id, 'span', self.start, self.end)

	@property
	def text(self):
		"""
		:rtype: str
		"""
		return self.obj.text


	@property
	def tokens(self):
		"""
		:rtype: list[Token]
		"""
		return self.document.tokens[self.start:self.end]







from .TokenSpan import TokenSpan
from .Entity import Entity
from .EntityType import EntityType
from .remove_list_duplicates import remove_list_duplicates
from .join_punctuation import join_punctuation



class EntityChunk(TokenSpan):
	def __init__(self, entity):
		"""
		:type entity: Entity
		"""
		super().__init__(obj=None, document=entity.document)
		self._start = min([entity.start]+[noun_chunk.start for noun_chunk in entity.parent_noun_chunks])
		self._end = max([entity.end]+[noun_chunk.end for noun_chunk in entity.parent_noun_chunks])
		self._entities = remove_list_duplicates([token.entity for token in self.tokens if token.entity])
		for token in self.tokens:
			token._entity_chunk = self


	@property
	def entities(self):
		"""
		:rtype: list[Entity]
		"""
		return self._entities

	@property
	def start(self):
		"""
		:rtype: int
		"""
		return self._start

	@property
	def end(self):
		"""
		:rtype: int
		"""
		return self._end

	@property
	def id(self):
		"""
		:rtype: tuple
		"""
		return (self.document.id, 'entity_chunk', self.start, self.end)

	def __str__(self):
		return ' '.join(join_punctuation(seq=[str(x) for x in self.tokens]))

	def __repr__(self):
		return str(self)

	@property
	def text(self):
		"""
		:rtype: str
		"""
		return str(self)

	@property
	def entity_types(self):
		"""
		:rtype: list[EntityType]
		"""
		return remove_list_duplicates([entity.entity_type for entity in self.entities])

	@property
	def main_entity_type(self):
		"""
		:rtype EntityType
		"""
		return sorted(self.entity_types)[0]

	def graph_str(self):
		"""
		:rtype: str
		"""
		return f"{self}\n({str(self.main_entity_type).replace('_', ' ')})"






PART_OF_SPEACH = {
	'adj': 'adjective',
	'adp': 'adposition',
	'adv': 'adverb',
	'cconj': 'coordinating_conjunction',
	'det': 'determiner',
	'propn': 'proper_noun',
	'punct': 'punctuation',
	'num': 'numeric_modifier',
	'sym': 'symbol',
	'pron':'pronoun'
}








class DocumentObject:
	def __init__(self, obj, document):
		"""
		:param obj: any object from a document
		:type document: Document
		"""
		self._obj = obj
		self._document = document

	@property
	def document(self):
		"""
		:rtype: Document
		"""
		return self._document

	def __repr__(self):
		return self._obj.__repr__()

	def __str__(self):
		return self._obj.__str__()

	def graph_str(self):
		return str(self)

	@property
	def id(self):
		return None

	def __eq__(self, other):
		return self.id==other.id

	def __ne__(self, other):
		return self.id!=other.id





from .TokenSpan import TokenSpan
from .remove_list_duplicates import remove_list_duplicates
from .EntityType import EntityType

class Entity(TokenSpan):

	def __init__(self, obj, document):
		super().__init__(obj=obj, document=document)
		for token in self.tokens:
			token._entity = self

	@property
	def name(self):
		return self._obj

	@property
	def parent_noun_chunks(self):
		return [nc for nc in self.document.noun_chunks if nc.start <= self.start and nc.end >= self.end]

	@property
	def child_noun_chunks(self):
		return [nc for nc in self.document.noun_chunks if nc.start >= self.start and nc.end <= self.end]

	@property
	def noun_chunks(self):
		return remove_list_duplicates(self.parent_noun_chunks + self.child_noun_chunks)


	@property
	def id(self):
		return (self.document.id, 'entity', self.start, self.end)

	def graph_str(self):
		return f"{self}\n({str(self.entity_type).replace('_', ' ')})"

	@property
	def entity_type(self):
		"""
		:rtype: EntityType
		"""
		return self.tokens[0].entity_type


from .NounChunk import NounChunk
from .Entity import Entity
from .BasicToken import BasicToken
from .EntityChunk import EntityChunk
from .EntityType import EntityType
from .remove_list_duplicates import remove_list_duplicates
from .join_punctuation import join_punctuation

from vineyard import Graph, NodeStyle
import spacy

nlp = spacy.load('en_core_web_sm')

class Token(BasicToken):
	def __add__(self, other):
		"""
		:param Token or Document other: adding two tokens (or several tokens) creates a Document
		:rtype: Document
		"""
		return Document(text=str(self).rstrip()+' '+str(other).lstrip())

	@property
	def parents(self):
		"""
		:rtype: list[Token]
		"""
		try:
			return self._parents
		except AttributeError:
			self._parents = [Token(obj=x, document=self.document) for x in self.token.ancestors]
			return self._parents

	@property
	def children(self):
		"""
		:rtype: list[Token]
		"""
		try:
			return self._children
		except AttributeError:
			self._children = [Token(obj=x, document=self.document) for x in self.token.children]
			return self._children



class Document:
	def __init__(self, text, id=0):
		"""
		:type text: str
		"""
		if isinstance(text, list):
			text = ' '.join([str(x) for x in text])
		self._id = id
		self._text = str(text)
		self._tokens = remove_list_duplicates([Token(obj=x, document=self) for x in self.doc])
		self._entities = remove_list_duplicates([Entity(obj=span, document=self) for span in self.doc.ents])
		self._noun_chunks = remove_list_duplicates([
			NounChunk(obj=span, document=self) for span in self.doc.noun_chunks
		])
		self._entity_chunks = remove_list_duplicates([EntityChunk(entity=entity) for entity in self.entities])

	@property
	def id(self):
		return self._id

	@property
	def doc(self):
		"""
		:rtype: tokens.Doc
		"""
		try:
			return self._doc
		except AttributeError:
			self._doc = nlp(self._text)
			return self._doc

	def __repr__(self):
		return self.doc.__repr__()

	def __str__(self):
		return self.doc.__str__()

	def graph_str(self):
		return f'd: {self}'[0:50]


	@property
	def tokens(self):
		"""
		:rtype: list[Token]
		"""
		return self._tokens

	@property
	def entities(self):
		"""
		:rtype: list[Entity]
		"""
		return self._entities

	@property
	def noun_chunks(self):
		"""
		:rtype: list[NounChunk]
		"""
		return self._noun_chunks

	@property
	def entity_chunks(self):
		"""
		:rtype: list[EntityChunk]
		"""
		return self._entity_chunks

	@property
	def entity_graph(self):
		"""
		:rtype: Graph
		"""
		try:
			return self._entity_graph
		except AttributeError:
			graph = Graph(ordering=False)

			# add tokens to the graph
			for token in self.tokens:
				graph.add_node(
					name=str(token.id), label=f'{token.graph_str()}',
					style=NodeStyle(text_size=8, shape='rect', style='"rounded, filled"')
				)

			# add noun_chunks to the graph
			for noun_chunk in self.noun_chunks:
				graph.add_node(
					name=str(noun_chunk.id), label=f'{noun_chunk.graph_str()}',
					style=NodeStyle(
						fill_colour='lightpink', text_colour='black', text_size=8,
						shape='rect', style='"rounded, filled"'
					)
				)

				# noun_chunk to its tokens
				for token in noun_chunk.tokens:
					graph.connect(start=str(token.noun_chunk.id), end=str(token.id))

			# add entities to the graph
			for entity in self.entities:
				graph.add_node(
					name=str(entity.id), label=f'{entity.graph_str()}',
					style=NodeStyle(fill_colour='gold3', text_colour='black')
				)
				# connect entity to its tokens
				for token in entity.tokens:
					graph.connect(start=str(token.id), end=str(entity.id))

			# add entity_chunks to the graph
			for entity_chunk in self.entity_chunks:
				graph.add_node(
					name=str(entity_chunk.id), label=f'{entity_chunk.graph_str()}',
					style=NodeStyle(fill_colour='gold', text_colour='black')
				)

				# connect entity_chunk to its entities
				for entity in entity_chunk.entities:
					graph.connect(start=str(entity.id), end=str(entity_chunk.id))

			self._entity_graph = graph
			return self._entity_graph

	@property
	def graph(self):
		"""
		:rtype: Graph
		"""
		try:
			return self._graph
		except AttributeError:
			graph = Graph(ordering=False)
			for token in self.tokens:
				graph.add_node(name=str(token.id), label=token.graph_str())
			for token in self.tokens:
				for child in token.children:
					dependency_label = str(child.dependency).replace('_', '\n')
					graph.connect(start=str(token.id), end=str(child.id), label=dependency_label)
			self._graph = graph
			return self._graph

	def __add__(self, other):
		"""
		:type other: Document or Token
		:rtype: Document
		"""
		return self.__class__(text=str(self).rstrip()+' '+str(other).lstrip())

	def bind(self, chunk_type='entity_chunk'):
		"""
		:param str chunk_type: can be one of entity_chunk, noun_chunk, or entity
		:rtype: list[Token or EntityChunk or NounChunk or Entity]
		"""
		if chunk_type=='entity_chunk':
			chunks = [x for x in self.tokens if x.entity_chunk is None] + self.entity_chunks
		elif chunk_type=='entity':
			chunks = [x for x in self.tokens if x.entity is None] + self.entities
		elif chunk_type=='noun_chunk':
			chunks = [x for x in self.tokens if x.noun_chunk is None] + self.noun_chunks
		else:
			raise ValueError(f'chunk_type "{chunk_type}" is not acceptable!')
		return sorted(chunks, key=lambda x: x.start)

	def mask_entity_types(self, entity_types, chunk_type='entity_chunk'):
		"""
		:param list[str] or list[EntityType] or str entity_types: entity types to be masked, it can be 'all'
		:param str chunk_type: can be one of entity_chunk, noun_chunk, or entity
		:rtype: str
		"""
		if entity_types=='all':
			pass
		else:
			entity_types = [EntityType(name=x) if not isinstance(x, EntityType) else x for x in entity_types]
		chunks = self.bind(chunk_type=chunk_type)
		strings = []
		for chunk in chunks:
			string = str(chunk)
			if 	(chunk_type=='entity_chunk' and isinstance(chunk, EntityChunk)) or \
				(chunk_type=='noun_chunk' and isinstance(chunk, NounChunk)):
				for entity_type in sorted(chunk.entity_types):
					if entity_types=='all':
						string = f'<{entity_type.long_name.upper()}>'
						break
					elif entity_type in entity_types:
						string = f'<{entity_type.long_name.upper()}>'
						break
			elif chunk_type=='entity' and isinstance(chunk, Entity):
				if entity_types=='all':
					string = f'<{chunk.entity_type.long_name.upper()}>'
				elif chunk.entity_type in entity_types:
					string = f'<{chunk.entity_type.long_name.upper()}>'
			strings.append(string)
		return ' '.join(join_punctuation(seq=[str(x) for x in strings]))






from setuptools import setup, find_packages


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='grammar',
	version='1.4',
	license='MIT',

	url='https://github.com/idin/grammar',
	author='Idin',
	author_email='py@idin.ca',

	description='Python library for natural language processing',
	long_description=readme(),
	long_description_content_type='text/markdown',

	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],

	packages=find_packages(exclude=["jupyter_tests", ".idea", ".git"]),
	install_requires=['vineyard'],
	python_requires='~=3.6',
	zip_safe=False
)


from collections import OrderedDict

# source: https://spacy.io/usage/linguistic-features
ENTITY_TYPES_BY_SHORTNAME = OrderedDict([
	('PERSON', ('People, including fictional', 'person')),
	('NORP', ('Nationalities or religious or political groups', 'group')),
	('FAC', ('Buildings, airports, highways, bridges, etc', 'facility')),
	('ORG', ('Companies, agencies, institutions, etc', 'organization')),
	('GPE', ('Countries, cities, states', 'geopolitical_entity')),
	('LOC', ('Non-GPE locations, mountain ranges, bodies of water', 'location')),
	('PRODUCT', ('Objects, vehicles, foods, etc (Not services)', 'product')),
	('EVENT', ('Named hurricanes, battles, wars, sports events, etc', 'event')),
	('WORK_OF_ART', ('Titles of books, songs, etc', 'word_of_art')),
	('LAW', ('Named documents made into laws', 'law')),
	('LANGUAGE', ('Any named language', 'language')),
	('DATE', ('Absolute or relative dates or periods', 'date')),
	('TIME', ('Times smaller than a day', 'time')),
	('PERCENT', ('Percentage, including "%"', 'percent')),
	('MONEY', ('Monetary values, including unit', 'money')),
	('QUANTITY', ('Measurements, as of weight or distance', 'quantity')),
	('ORDINAL', ('"first", "second", etc', 'ordinal')),
	('CARDINAL', ('Numerals that do not fall under another type', 'cardinal'))
])
ENTITY_TYPES_BY_LONGNAME = OrderedDict()
for shortname, definition_longname in ENTITY_TYPES_BY_SHORTNAME.items():
	definition, longname = definition_longname
	ENTITY_TYPES_BY_LONGNAME[longname] = (definition, shortname)

class EntityType:
	def __init__(self, name):
		name_upper = str(name).upper()
		name_lower = str(name).lower()
		self._short_name = name_upper
		self._long_name = name_lower
		self._description = name
		if name_upper in ENTITY_TYPES_BY_SHORTNAME:
			entity_description_and_long_name = ENTITY_TYPES_BY_SHORTNAME[name_upper]
			self._description = entity_description_and_long_name[0]
			self._long_name = entity_description_and_long_name[1].lower()
			self._index = list(ENTITY_TYPES_BY_SHORTNAME.keys()).index(name_upper)
		elif name_lower in ENTITY_TYPES_BY_LONGNAME:
			entity_description_and_long_name = ENTITY_TYPES_BY_LONGNAME[name_lower]
			self._description = entity_description_and_long_name[0]
			self._short_name = entity_description_and_long_name[1]
			self._index = list(ENTITY_TYPES_BY_LONGNAME.keys()).index(name_lower)

	@property
	def short_name(self):
		return self._short_name

	@property
	def description(self):
		return self._description

	@property
	def long_name(self):
		return self._long_name

	@property
	def index(self):
		return self._index

	@property
	def id(self):
		return (self.index, self.short_name)

	def __str__(self):
		return self.long_name

	def __repr__(self):
		return self.long_name

	def __lt__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id < other.id

	def __eq__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id == other.id

	def __gt__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id > other.id

	def __le__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id <= other.id

	def __ge__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id >= other.id

	def __ne__(self, other):
		"""
		:type other: EntityType
		:rtype: bool
		"""
		return self.id != other.id



